@@ -368,66 +368,6 @@ static void bsp_determine_snp(struct cpuinfo_x86 *c)
 #endif
 }
 
-static bool amd_check_tsa_microcode(void)
-{
-	struct cpuinfo_x86 *c = &boot_cpu_data;
-	union zen_patch_rev p;
-	u32 min_rev = 0;
-
-	p.ext_fam	= c->x86 - 0xf;
-	p.model		= c->x86_model;
-	p.ext_model	= c->x86_model >> 4;
-	p.stepping	= c->x86_stepping;
-	/* reserved bits are expected to be 0 in test below */
-	p.__reserved	= 0;
-
-	if (cpu_has(c, X86_FEATURE_ZEN3) ||
-	    cpu_has(c, X86_FEATURE_ZEN4)) {
-		switch (p.ucode_rev >> 8) {
-		case 0xa0011:	min_rev = 0x0a0011d7; break;
-		case 0xa0012:	min_rev = 0x0a00123b; break;
-		case 0xa0082:	min_rev = 0x0a00820d; break;
-		case 0xa1011:	min_rev = 0x0a10114c; break;
-		case 0xa1012:	min_rev = 0x0a10124c; break;
-		case 0xa1081:	min_rev = 0x0a108109; break;
-		case 0xa2010:	min_rev = 0x0a20102e; break;
-		case 0xa2012:	min_rev = 0x0a201211; break;
-		case 0xa4041:	min_rev = 0x0a404108; break;
-		case 0xa5000:	min_rev = 0x0a500012; break;
-		case 0xa6012:	min_rev = 0x0a60120a; break;
-		case 0xa7041:	min_rev = 0x0a704108; break;
-		case 0xa7052:	min_rev = 0x0a705208; break;
-		case 0xa7080:	min_rev = 0x0a708008; break;
-		case 0xa70c0:	min_rev = 0x0a70c008; break;
-		case 0xaa002:	min_rev = 0x0aa00216; break;
-		default:
-			pr_debug("%s: ucode_rev: 0x%x, current revision: 0x%x\n",
-				 __func__, p.ucode_rev, c->microcode);
-			return false;
-		}
-	}
-
-	if (!min_rev)
-		return false;
-
-	return c->microcode >= min_rev;
-}
-
-static void tsa_init(struct cpuinfo_x86 *c)
-{
-	if (cpu_has(c, X86_FEATURE_HYPERVISOR))
-		return;
-
-	if (cpu_has(c, X86_FEATURE_ZEN3) ||
-	    cpu_has(c, X86_FEATURE_ZEN4)) {
-		if (amd_check_tsa_microcode())
-			setup_force_cpu_cap(X86_FEATURE_VERW_CLEAR);
-	} else {
-		setup_force_cpu_cap(X86_FEATURE_TSA_SQ_NO);
-		setup_force_cpu_cap(X86_FEATURE_TSA_L1_NO);
-	}
-}
-
 static void bsp_init_amd(struct cpuinfo_x86 *c)
 {
 	if (cpu_has(c, X86_FEATURE_CONSTANT_TSC)) {
@@ -535,9 +475,6 @@ static void bsp_init_amd(struct cpuinfo_x86 *c)
 	}
 
 	bsp_determine_snp(c);
-
-	tsa_init(c);
-
 	return;
 
 warn:
@@ -925,16 +862,6 @@ static void init_amd_zen1(struct cpuinfo_x86 *c)
 
 	pr_notice_once("AMD Zen1 DIV0 bug detected. Disable SMT for full protection.\n");
 	setup_force_cpu_bug(X86_BUG_DIV0);
-
-	/*
-	 * Turn off the Instructions Retired free counter on machines that are
-	 * susceptible to erratum #1054 "Instructions Retired Performance
-	 * Counter May Be Inaccurate".
-	 */
-	if (c->x86_model < 0x30) {
-		msr_clear_bit(MSR_K7_HWCR, MSR_K7_HWCR_IRPERF_EN_BIT);
-		clear_cpu_cap(c, X86_FEATURE_IRPERF);
-	}
 }
 
 static bool cpu_has_zenbleed_microcode(void)
@@ -979,13 +906,6 @@ static void init_amd_zen2(struct cpuinfo_x86 *c)
 	init_spectral_chicken(c);
 	fix_erratum_1386(c);
 	zen2_zenbleed_check(c);
-
-	/* Disable RDSEED on AMD Cyan Skillfish because of an error. */
-	if (c->x86_model == 0x47 && c->x86_stepping == 0x0) {
-		clear_cpu_cap(c, X86_FEATURE_RDSEED);
-		msr_clear_bit(MSR_AMD64_CPUID_FN_7, 18);
-		pr_emerg("RDSEED is not reliable on this platform; disabling.\n");
-	}
 }
 
 static void init_amd_zen3(struct cpuinfo_x86 *c)
@@ -1018,43 +938,8 @@ static void init_amd_zen4(struct cpuinfo_x86 *c)
 	}
 }
 
-static bool check_rdseed_microcode(void)
-{
-	struct cpuinfo_x86 *c = &boot_cpu_data;
-	union zen_patch_rev p;
-	u32 min_rev = 0;
-
-	p.ext_fam	= c->x86 - 0xf;
-	p.model		= c->x86_model;
-	p.ext_model	= c->x86_model >> 4;
-	p.stepping	= c->x86_stepping;
-	/* reserved bits are expected to be 0 in test below */
-	p.__reserved	= 0;
-
-	if (cpu_has(c, X86_FEATURE_ZEN5)) {
-		switch (p.ucode_rev >> 8) {
-		case 0xb0021:	min_rev = 0xb00215a; break;
-		case 0xb1010:	min_rev = 0xb101054; break;
-		default:
-			pr_debug("%s: ucode_rev: 0x%x, current revision: 0x%x\n",
-				 __func__, p.ucode_rev, c->microcode);
-			return false;
-		}
-	}
-
-	if (!min_rev)
-		return false;
-
-	return c->microcode >= min_rev;
-}
-
 static void init_amd_zen5(struct cpuinfo_x86 *c)
 {
-	if (!check_rdseed_microcode()) {
-		clear_cpu_cap(c, X86_FEATURE_RDSEED);
-		msr_clear_bit(MSR_AMD64_CPUID_FN_7, 18);
-		pr_emerg_once("RDSEED32 is broken. Disabling the corresponding CPUID bit.\n");
-	}
 }
 
 static void init_amd(struct cpuinfo_x86 *c)
@@ -1160,8 +1045,13 @@ static void init_amd(struct cpuinfo_x86 *c)
 	if (!cpu_feature_enabled(X86_FEATURE_XENPV))
 		set_cpu_bug(c, X86_BUG_SYSRET_SS_ATTRS);
 
-	/* Enable the Instructions Retired free counter */
-	if (cpu_has(c, X86_FEATURE_IRPERF))
+	/*
+	 * Turn on the Instructions Retired free counter on machines not
+	 * susceptible to erratum #1054 "Instructions Retired Performance
+	 * Counter May Be Inaccurate".
+	 */
+	if (cpu_has(c, X86_FEATURE_IRPERF) &&
+	    (boot_cpu_has(X86_FEATURE_ZEN1) && c->x86_model > 0x2f))
 		msr_set_bit(MSR_K7_HWCR, MSR_K7_HWCR_IRPERF_EN_BIT);
 
 	check_null_seg_clears_base(c);
